/*
 *                      ..::PictoGame::..
 *
 * Copyright (C) Federico Vera 2010-2011 <dktcoding [at] gmail>
 *
 * PictoGame is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or any later
 * version.
 *
 * PictoGame is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package picto.utils;

import java.util.Arrays;
import java.util.HashMap;
import picto.app.interfaces.Flushable;
import picto.app.interfaces.Status;
import picto.exceptions.IncorrectHeaderException;

/**
 * I tried a <b>LOT</b> of different approaches to solve the F****** nonograms
 * and this is the best I could come up with. The basic idea is to generate all
 * of the possible solutions, overlap them, and mark all of the cells that
 * are in the intersection as FULL, after that the process is repeated with the
 * binary complement of the solutions and all the cells in the intersection are
 * marked as EMPTY. This process is repeated in all the rows and the columns.
 * <br>Once this is done all the solutions are compared against the data in the
 * board, if a solution doesn't fit (if a cell is marked as FULL or EMPTY when
 * the board specifies otherwise) it's discarded.
 * <br>All of this is repeated until there are no more UNKNOWN cells in the
 * board or until a loop ends without any changes being made in the board.-
 *
 * <br>The algorithm is fairly quick and doesn't have a big memory heap (at
 * least compared to others that I've seen on the web), but it still needs a lot
 * of fine tuning. If anyone has a suggestion on how to improve this, or how to
 * create a better one I'll be very happy to hear about it.
 * <br><br> Limitations: <ol>
 * <li> The maximum row and column length is 64 cells, this is because solutions
 *      are treated as <code>long</code> numbers, this can be easily remedied by
 *      using <code>BigIntegers</code> of course this means that you'll be
 *      sacrificing a bit of performance and memory... in fact the whole idea of
 *      using numbers as solutions is lost. <s><i>Perhaps <code>Strings</code>
 *      would be faster for this?</i></s> NO!
 * <li> The solver can only be used with monochromatic nonograms but I think
 *      that it could be easily solved doing the following:<ol>
 *          <li> Remove the inner space between segments when a solution is
 *               calculated (color nonograms don't require the segments to be
 *               separated).
 *          <li> Since the only REAL difference between the monochromatic and
 *               the multicolor nonograms is the one specified above, the only
 *               extra step needed to solve them is that once a line is marked
 *               as complete you must "<i>paint</i>" the cells according to the
 *               header (and that seems <i>easy</i> enough).</ol>
 * <li> If the nonogram contains more than one possible solution it's probable
 *      that the solver will fail, not every time though, because if an
 *      unambiguous solution can be found in a first try then the solver will
 *      guess it... I'm not sure why this happens (statistics were never my
 *      strong side), but I'm working on it!
 * </ol><br> TODO list: <ol>
 * <li><s>Remove all those <code>Long.parseLong()</code>, they are pretty slow
 *      and require the creation of a lot of <code>String</code> objects...
 *      there must be a better way.</s>
 * <li> Add support for multiple solutions, since all of the possibilities are
 *      already calculated there must be a simple way of doing it.
 * <li><s>Add support for button drawing, it would be nice to integrate the
 *      solver with the game without loosing the possibility of using it
 *      separately.</s>
 * <li> General tuning up.
 * </ol>
 * <br><i>Note: </i>Some people pointed out that I use an awful lot of bitwise
 * operations even when they are not needed, and that is correct! I'm not a
 * <tt>"bit wizard"</tt> but at least in my mind the best way to understand
 * something is to play with it, so that's what I've done!
 * <br><i>Other note: </i>I can't believe that someone actually read the code =P
 * <br><br>I wanted to thank Javier for contributing the method that calculate
 * the number of combinations! Now there's no need for ArrayLists!
 *
 * @version 0.5
 * @author Fido <dktcoding [at] gmail>
 */
public class PictoSolver implements Flushable{
    
///////////////////////////////////////////////////
    /**Indicates that the line is a row*/
    private static final byte ROW     = 0;
    /**Indicates that the line is a column*/
    private static final byte COLUMN  = 1;
    /**Indicates that the cell is full*/
    private static final byte FULL    = 2;
    /**Indicates that the cell is empty*/
    private static final byte EMPTY   = 3;
    /**Indicates that the cell status is unknown*/
    private static final byte UNKNOWN = 4;
///////////////////////////////////////////////////

    /**This array contains the board*/
    private static byte[][] cells;
    /**
     * This contains the column headers.<br>
     * <i>Note: </i> the <code>calculateEasySolution(byte idx, byte type)</code>
     * method used to remove some elements from here, but since it's no longer
     * used the headers should stay there
     * @see calculateEasySolution(byte, byte)
     */
    private byte[][] columnHeaders;
    /**
     * This contains the row headers.<br>
     * <i>Note: </i> the <code>calculateEasySolution(byte idx, byte type)</code>
     * method used to remove some elements from here, but since it's no longer
     * used the headers should stay there
     * @see calculateEasySolution(byte, byte)
     */
    private byte[][] rowHeaders;
    /**
     * Number of rows in the board
     */
    private byte num_rows;
    /**
     * Number of columns in the board
     */
    private byte num_columns;
    /**
     * Number of row headers
     */
    private byte num_row_headers;
    /**
     * Number of column headers
     */
    private byte num_column_headers;
    /**
     * This array list contains all the column solutions
     */
    private Solution[][] columnSolutions;
    /**
     * This array list contains all the row solutions
     */
    private Solution[][] rowSolutions;

    private ColoredButton[][] buttonBoard;

    /**
     * Generates a new solver.
     *
     * @param ps non-flushed PhraseScrambler
     */
    public PictoSolver(final PhraseScrambler ps){
        this(ps.getPic());
    }

    /**
     * Generates a new solver.
     *
     * @param pic non-flushed Picto
     */
    public PictoSolver(final Picto pic){
        num_rows = pic.getNumberOfRows();
        num_columns = pic.getNumberOfColumns();
        num_row_headers = pic.getRowHeaderNum();
        num_column_headers = pic.getColumnHeaderNum();
        rowHeaders = new byte[num_rows][num_row_headers];
        columnHeaders = new byte[num_columns][num_column_headers];
        
        for (byte i = 0; i < num_rows; i++){
            for (byte j = 0; j < num_row_headers; j++){
                rowHeaders[i][j] = pic.getRowHeadersElement(i, j);
            }
        }
        
        for (byte i = 0; i < num_columns; i++){
            for (byte j = 0; j < num_column_headers; j++){
                columnHeaders[i][j] = pic.getColumnHeadersElement(i, j);
            }
        }
        
        columnSolutions = new Solution[num_columns][];
        rowSolutions = new Solution[num_rows][];
        generateBoard();
    }

    /**
     * Generates a new solver.
     *
     * @param columnHeader
     * @param rowHeader
     */
    public PictoSolver(final byte[][] columnHeader, final byte[][] rowHeader){
        rowHeaders = rowHeader;
        columnHeaders = columnHeader;
        num_column_headers = (byte)columnHeader[0].length;
        num_row_headers = (byte)rowHeader[0].length;
        num_columns = (byte)columnHeader.length;
        num_rows = (byte)rowHeader.length;
        columnSolutions = new Solution[num_columns][];
        rowSolutions = new Solution[num_rows][];
        generateBoard();
    }

    private HashMap<String, Solution[]> solutions_map;
    public void solve() throws IncorrectHeaderException{
        byte i; solutions_map = new HashMap<String, Solution[]>(num_columns + num_rows);
        
        for(i = 0; i < num_columns; i++){
            if (solutions_map.containsKey(Arrays.toString(columnHeaders[i]))){
                Solution[] temp = solutions_map.get(Arrays.toString(columnHeaders[i]));
                columnSolutions[i] = new Solution[temp.length];
                System.arraycopy(temp, 0, columnSolutions[i], 0, temp.length);
            } else {
                columnSolutions[i] = generateAllPosibleSolutions(columnHeaders[i], num_rows, getInitialIndex(columnHeaders[i]));
                solutions_map.put(Arrays.toString(columnHeaders[i]), columnSolutions[i]);
            }
        }
        
        if (num_rows != num_columns){
            solutions_map.clear();
        }
        
        for(i = 0; i < num_rows; i++){
            if (solutions_map.containsKey(Arrays.toString(rowHeaders[i]))){
                Solution[] temp = solutions_map.get(Arrays.toString(rowHeaders[i]));
                rowSolutions[i] = new Solution[temp.length];
                System.arraycopy(temp, 0, rowSolutions[i], 0, temp.length);
            } else {
                rowSolutions[i] = generateAllPosibleSolutions(rowHeaders[i], num_columns, getInitialIndex(rowHeaders[i]));
                solutions_map.put(Arrays.toString(rowHeaders[i]), rowSolutions[i]);
            }
        }
        solutions_map.clear();
        long[] counter, counter2 = countNumberOfSolutions();
        do{
            counter = counter2;
            
            for(i = 0; i < num_rows; i++){
                setOnes(overlapSolutions(rowSolutions[i]), i, ROW, FULL);
                setOnes(overlapRevertedSolutions(rowSolutions[i]), i, ROW, EMPTY);
            }
            
            for(i = 0; i < num_columns; i++){
                removeImpossibleSolutions(i, COLUMN);
            }
            
            for(i = 0; i < num_columns; i++){
                setOnes(overlapSolutions(columnSolutions[i]), i, COLUMN, FULL);
                setOnes(overlapRevertedSolutions(columnSolutions[i]), i, COLUMN, EMPTY);
            }
            
            for(i = 0; i < num_rows; i++){
                removeImpossibleSolutions(i, ROW);
            }
            
            counter2 = countNumberOfSolutions();
        } while ((counter[0] != counter2[0] || counter[1] != counter2[1]) && !isSolved());
    }

    /**
     * This method is used for debugging, it tells how many solutions are
     * available.
     *
     * @return A long array containing the number of solutions available, the
     *         indexes are the same that those specified as flags
     *         <code>(<b>COLUMN|ROW</b>)</code>
     */
    private long[] countNumberOfSolutions(){
        long row_counter = 0, column_counter = 0;
        
        for (Solution[] row_sol : rowSolutions){
            row_counter += (row_sol == null) ? 0 : row_sol.length;
        }
        
        for (Solution[] col_sol : columnSolutions){
            column_counter += (col_sol == null) ? 0 : col_sol.length;
        }
        
        return new long[]{row_counter, column_counter};
    }

    /**
     * Initializes a new board with all elements <tt>UNKNOWN</tt>
     */
    private void generateBoard(){
        cells = new byte [num_rows][num_columns];
        for (byte[] row : cells){
            Arrays.fill(row, UNKNOWN);
        }
    }
    
    /**
     * Calculates the minimum line size necessary to fit the segments specified
     * in the header.
     *
     * @param header_line 
     * @param initial_idx index in which you want to start counting
     * @return minimum size needed to fit the segments in the line
     */
    private static byte getMinimumLineSize(final byte[] header_line,
                                           final byte initial_idx){
        byte how_many_numbers = 0, counter = 0;
        
        for (byte i = initial_idx, n = (byte)header_line.length; i < n; i++){
            if (header_line[i] != 0){
                counter += header_line[i];
                how_many_numbers++;
            }
        }
        
        if (how_many_numbers != 0){
            counter += how_many_numbers - 1;
        }
        
        return counter;
    }
    
    /**
     * Returns the index of the first non-zero header
     * 
     * @param header_line
     * @return index of the first non-zero element or <tt>-1</tt> if all 
     *         elements are zero
     */
    private static byte getInitialIndex(final byte[] header_line){
        for (byte i = 0, n = (byte)header_line.length; i < n; i++){
            if (header_line[i] != 0){
                return i;
            }
        }
        return -1;
    }

    /**
     * This method sets the data in the line, note that this method only sets
     * the cells that are confirmed to be <tt>EMPTY</tt>.
     *
     * @param data data as specified in <code>Picto</code>
     * @param type type of line <code>(<b>COLUMN|ROW</b>)</code>
     * @param idx data index
     */
    private void setOnes (long data, final byte idx, final byte type, final byte as){
        assert (type == ROW || type == COLUMN)
                     : "setOnesAsEmpty was called with an invalid type of line!"
                     + "\n WTF were you thinking!?";
        if (data == 0L){
            return;
        }
        data &= (1L << ((type == ROW) ? num_columns : num_rows)) - 1L;

        // Sorry for this but both emptys and fulls have different values...
        byte status = as == EMPTY ? Status.EMPTY : Status.FULL;
        boolean control_panel = buttonBoard != null;

        if (type == COLUMN) {
            long mask = 1L;
            byte i = num_rows;
            while (--i >= 0) {
                if ((data & mask) != 0) {
                    cells[i][idx] = as;
                    if (control_panel) {
                        buttonBoard[i][idx].setStatus(status);
                    }
                }
                mask <<= 1;
            }
        } else {
            long mask = 1L;
            byte i = num_columns;
            while (--i >= 0) {
                if ((data & mask) != 0) {
                    cells[idx][i] = as;
                    if (control_panel) {
                        buttonBoard[idx][i].setStatus(status);
                    }
                }
                mask <<= 1;
            }
        }
    }
    
    /**
     * This method checks if there is any <tt>UNKNOWN</tt> cell in order to 
     * check if it's solved
     * 
     * @return <code>true</code> if there is no <tt>UNKNOWN</tt> cell and 
     *         <code>false</code> if it has some <tt>UNKNOWN</tt> cell
     */
    private boolean isSolved(){
        for (byte[] row : cells){
            for (byte element : row){
                if (element == UNKNOWN){
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Return a string repeated 'multiplier' times.
     *
     * @param str String you want to multiply
     * @param n how many times you want to multiply it
     * @return str n times
     * @throws IllegalArgumentException if multiplier is negative
     */
    private String generateStringBlock(final String str, final int n){
        if (n < 0){
            throw new IllegalArgumentException ("Multiplier must be a positive integer!");
        }
        
        StringBuilder temp = new StringBuilder(str.length() * n);
        for (int i = 0; i < n; i++){
            temp.append(str);
        }
        
        return temp.toString();
    }

    /**
     * This method generates all the possible solutions for a line
     * It should always return at least <b>one</b> solution.
     *
     * @param headers the header that will be used to calculate the solutions
     * @param size the size (num_rows|num_columns)
     * @param start_pt in which of the header elements to begin
     * @return ArrayList of <code>Solution</code> containing all the possible
     *         solutions
     * @throws OutOfMemoryError like... <b>a lot!</b>
     * @see Picto
     * @see Solution
     */
    private Solution[] generateAllPosibleSolutions(final byte[] headers,
                                                   final byte size,
                                                   final byte start_pt){
        if (start_pt == -1){
            return new Solution[]{new Solution()};
        }
        byte i = (byte)(size - getMinimumLineSize(headers, (byte)(start_pt + 1)) - 1);
        Solution[] sols = new Solution[calculateNumberOfCombinations(headers, size, start_pt)];
        if (headers.length-1 == start_pt){
            int right_offset = size - headers[start_pt] + 1;
            Solution foo = new Solution(1L);
            // Generate a ones segment
            foo.solution <<= headers[start_pt];
            foo.solution--;
            // "Move" the segment
            while (right_offset > 0){
                sols[--right_offset] = new Solution(foo.solution << right_offset, foo.offset++);
            }
        } else {
            Solution num = new Solution(1L);
            // Generate a ones segment with a trailing zero
            num.solution <<= headers[start_pt] + 1;
            num.solution -= 2;
            for (int j = headers[start_pt], k = 0; j <= i ; j++){
                Solution[] foo = generateAllPosibleSolutions
                    (headers, (byte)(size - j - 1 ), (byte)(start_pt + 1));
                for (Solution part_sol : foo){
                    // Append the previous solution to the current one
                    part_sol.solution |= num.solution << (64 + part_sol.offset
                                 - Long.numberOfLeadingZeros(part_sol.solution));
                    part_sol.offset = num.offset;
                    sols[k++] = part_sol;
                }
                // Add an extra leading zero
                num.offset++;
            }
        }
        return sols;
    }

    ///////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////

    /**
     * Calculate the number of possible combinations (solutions)
     *
     * @param header the header of the line
     * @param size the lines's length
     * @param start_pt the initial index within the header
     * @return the number of possible combinations
     * @author Javier Laje <javier_laje [at] hotmail>
     */
    private static int calculateNumberOfCombinations(final byte[] header,
                                                     final byte size,
                                                     final byte start_pt){
        if (header.length -1 == start_pt){
            return size - header[header.length-1] + 1;
        }
        byte TiT_able_column = (byte)(header.length - start_pt - 2);
        byte TI = calculateTIFactor(header, size, start_pt);
        int num_of_posibilities = 0;
        for (int i = 0; i < TI; i++){
            num_of_posibilities += TiT_able[i][TiT_able_column] * (TI - i);
        }
        return num_of_posibilities;
    }


    private static byte calculateTIFactor(final byte[] header,
                                          final byte size,
                                          final byte start_pt){
        byte ti_factor = (byte)(size + 2);
        for (byte i = start_pt; i < header.length; i++){
            ti_factor -= header[i];
        }
        ti_factor -= (header.length - start_pt);
        return ti_factor;
    }

    /**
     * Array containing Tartaglia's Triangle
     */
    private static Long[][] TiT_able = fillTartagliasTriangle();

    /**
     * This method generates the classic Tartaglia's Triangle
     *
     * @TODO is not necessary to generate all the elements on the list (is symmetric)
     * @TODO perhaps change it for the formula? n!/(m!(n-m)!)
     * @author Javier Laje <javier_laje [at] hotmail>
     */
    private static Long[][] fillTartagliasTriangle(){
        Long[][] TiT = new Long[Long.SIZE][Long.SIZE];
        Arrays.fill(TiT[0], 1l);
        for (int i = 1; i < Long.SIZE; i++){
            TiT[i][0] = 1l;
            for (int j = 1; j < Long.SIZE; j++){
                TiT[i][j] = TiT[i-1][j] + TiT[i][j-1];
            }
        }
        return TiT;
    }

    ///////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////

    /**
     * This class tries to avoid all the string generation...<br>
     * Only time will tell if it manages to reduce the heap and the generation 
     * time of <code>generateAllPosibleSolutions(byte[], byte, byte)</code><br>
     * Time has spoken and it really improve performance!<br>
     * Although that extra byte used for each solution is adding a lot of extra
     * fu*****g memory... I'll solve it eventually... I have some real job now
     */
    private class Solution implements Cloneable{
        /**Contains the solution*/
        public long solution;
        /**Contains the number of leading zeros*/
        public byte offset;

        private Solution(){
            this(0L, (byte)0);
        }

        private Solution(final long solution){
            this(solution, (byte)0);
        }

        private Solution(final long solution, final byte offset){
            this.solution = solution;
            this.offset = offset;
        }

        @Override 
        public Solution clone(){
            return new Solution(solution, offset);
        }

        @Override 
        public boolean equals(final Object obj){
            if (obj instanceof Solution){
                Solution temp = (Solution) obj;
                return temp.offset == offset && temp.solution == solution;
            }
            return false;
        }

        @Override 
        public int hashCode() {
            int hash = 5;
            hash = 73 * hash + (int) (this.solution ^ (this.solution >>> 32));
            hash = 73 * hash + this.offset;
            return hash;
        }

        @Override
        public String toString(){
            return generateStringBlock("0", offset)
                    + Long.toBinaryString(solution);
        }
    }

    /**
     * This method computes those cells that are <b>DEFINITELY</b> <tt>FULL<tt>
     * according to the solutions passed
     *
     * @param sols ArrayList containing the possible solutions
     * @return a <code>long</code> number containing the overlapped solution,
     *         that means, a number that contains ones in the position where
     *         the cells are marked as <tt>FULL</tt>. If the <code>sols</code>
     *         array is empty then it returns 0.
     * @see generateAllPosibleSolutions(byte[], byte, byte)
     */
    private long overlapSolutions(final Solution[] sols){
        if (sols[0] == null){
            return 0l;
        }
        
        long foo = sols[0].solution;
        int i = sols.length;
        
        while (i > 1){
            if (foo == 0){
                break;
            }
            foo &= sols[--i].solution;
        }
        
        return foo;
    }

    /**
     * This method computes those cells that are <b>DEFINITELY</b> <tt>EMPTY<tt>
     * according to the solutions passed
     *
     * @param sols ArrayList containing the possible solutions
     * @return a <code>long</code> number containing the overlapped solution,
     *         that means, a number that contains ones in the position where
     *         the cells are marked as <tt>EMPTY</tt>. If the <code>sols</code>
     *         array is empty then it returns 0.
     * @see generateAllPosibleSolutions(byte[], byte, byte)
     */
    private long overlapRevertedSolutions(final Solution[] sols){
        if (sols[0] == null){
            return 0l;
        }
        
        long foo = ~sols[0].solution;
        int i = sols.length;
        
        while (i > 1){
            if (foo == 0){
                break;
            }
            foo &= ~sols[--i].solution;
        }
        
        return foo;
    }

    /**
     * This method checks and removes all the solutions that are definitely
     * wrong.<br>
     * If there's a line that is complete then it will remove all the solutions,
     * including the one that is used, that way it won't be checked the next
     * time.
     *
     * @param idx Index of the line
     * @param type type of line <code>(<b>COLUMN|ROW</b>)</code>
     * @return <code>true</code> if some solution was removed and
     *         <code>false</code> if none was removed.
     */
    private void removeImpossibleSolutions(final byte idx, final byte type){
        assert (type == ROW || type == COLUMN)
                     : "removeImpossibleSolutions was called with an invalid "
                     + "type of line!\nWTF were you thinking!?";
        Solution[] foo = (type == ROW) ? rowSolutions[idx]
                                       : columnSolutions[idx];
        // Calculate the current data in the specified line
        long curr_full_cells = 0, curr_empty_cells = 0;
        
        if (type == ROW){
            for (int i = 0; i < num_columns; i++){
                curr_full_cells  <<= 1;
                curr_empty_cells <<= 1;
                if (cells[idx][i] == FULL){
                    curr_full_cells |= 1;
                } else if (cells[idx][i] == EMPTY){
                    curr_empty_cells |= 1;
                }
            }
        } else {
            for (int i = 0; i < num_rows; i++){
                curr_full_cells  <<= 1;
                curr_empty_cells <<= 1;
                if (cells[i][idx] == FULL){
                    curr_full_cells |= 1;
                } else if (cells[i][idx] == EMPTY){
                    curr_empty_cells |= 1;
                }
            }
        }
        
        // Remove all the elements that contradict the current board
        if ((curr_full_cells|curr_empty_cells) == 0){
            return;
        }
        
        int counter = 0;
        for (int i = 0, n = foo.length; i < n; i++){
            if ((( foo[i].solution & curr_full_cells ) != curr_full_cells
               ||(~foo[i].solution & curr_empty_cells) != curr_empty_cells)){
                foo[i] = null;
                counter++;
            }
        }
        
        if (counter != 0){
            Solution[] temp = new Solution[foo.length-counter];
            int i = 0, j = 0, n = temp.length;
            
            while (i < n){
                if (foo[j] != null){
                    temp[i] = foo[j];
                } else {
                    i--;
                }
                j++; i++;
            }
            
            if (type == ROW){
                rowSolutions[idx] = temp;
            } else {
                columnSolutions[idx] = temp;
            }
        }
    }

    /**
     * Sets the board. This is done for PictoGame's sake and not the solver's,
     * the idea is that if a board is set then the solution is also shown in the
     * playing board.
     *
     * @param board board in which to draw the solution.
     */
    public void setButtonBoard(final ColoredButton[][] board){
        this.buttonBoard = board;
    }

    @Override public void flush(){
        buttonBoard = null;
        rowSolutions = null;
        columnSolutions = null;
    }

    public static void main(String ... args) throws IncorrectHeaderException{
        // Example data
//        byte[][] COLUMNS =
//        {{3},{5},{7},{7},{7}
//        ,{5},{3},{2},{2},{2}
//        ,{2,1},{2,4},{1,1,2,4},{5,7},{2,2,6}
//        ,{11},{13},{4,7},{1,2,2},{2,2}
//        ,{2,3},{3},{1}};
//        byte[][] ROWS =
//        {{5},{7},{1,3},{6},{4}
//        ,{2},{12},{15},{3,2,5,1},{5,5,2}
//        ,{7,5},{7,4,5},{7,3,5},{5,2,1},{3,3}};
        byte[][] ROWS = {{2,2},
        {18},{1,1},{1,3,1,1,3,1},{1,3,2,2,3,1},{1,2,5,3,1},{1,2,5,1,1,1},
        {1,3,1,1,1,1,1,1},{1,3,1,1,3,1},{1,2,1,1,3,1},{2,2,1,1,1,4,2},
        {2,3,1,1,1,14},{2,2,1,8,1},{5,5,1,1},{4,4,1,2,1,1,1},{6,1,2,1,1,1},
        {4,1,1,1,1,1},{2,1,1,1,1,1},{2,1,2,1,1,1},{2,1,2,1,1,1},
        {2,1,1,1,1,1},{2,1,1,1,1,1},{0,2,1,1,3,1},{2,1,1,3,1},
        {2,2,2},{2,2,2},{2,2,2},{4,3,3},{10,8},{14,8},{14,6},{4},{2},{2},{2}};

        byte[][] COLUMNS = {{12},{2,4},{1,2,2},{1,9,2,2},{1,9,2,3},{1,2,2,1,2,3},
           {1,2,3},{1,9,2,4},{1,3,16},{1,3,16},{1,3,2,4},{1,9,2,3},{1,2,3},
           {1,9,2,3},{1,3,2,2,2},{1,7,2,2},{2,4},{12},{16},{2,4},{2,4},{2,10,4},
           {2,2,2,4},{2,7},{2,10,7},{2,2,4},{2,10,4},{2,4},{2,4},{16}};

        PictoSolver ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/clubs.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(COLUMNS, ROWS);
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/bow.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/gnome.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/blender.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/castle_tower.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/yin_yang.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/coat_of_arms.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/floppy.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/kde.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/ubuntu.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/daphne.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/bridge.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/bulldozer.pattern"), false));
        ps.solve();
        System.out.println(ps);
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/flowers.pattern"), false));
        ps.solve();
        System.out.println(ps);
        System.out.println("Este es jodido!");
        ps = new PictoSolver(new Picto(PictoSolver.class.getResourceAsStream("/res/patterns/chalice.pattern"), false));
        ps.solve();
        System.out.println(ps);
    }

    @Override public String toString() {
        StringBuilder str = new StringBuilder((cells.length + 1) * cells[0].length + 30);
        
        str.append("PictoSolver (")
           .append(num_rows)
           .append('x')
           .append(num_columns)
           .append(')').append('\n');
        
        for (byte[] line : cells){
            
            for (byte element : line){
                
                switch (element){
                    case UNKNOWN:
                        str.append('?');
                        break;
                    case FULL:
                        str.append('*');
                        break;
                    default:
                        str.append(' ');
                }
                
            }
            
            str.append('\n');
        }
        
        return str.toString();
    }

}